// Copyright 1997-1999 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.


#define RCS_ID(rcsIdString) \
        static void *rcs_id = rcsIdString; \
        static void *__rcs_id_hack() { __rcs_id_hack(); return rcs_id; }

RCS_ID("$Header: /Network/Source/CVS/OmniGroup/Bundles/EOModeler/OmniCVSEOModelerBundle/CVSManager.m,v 1.10 1999/09/30 21:41:14 andrew Exp $")


#import <Foundation/Foundation.h>
#import <EOModeler/EOModeler.h>

#import "CVSManager.h"


/* minor extension to NSFileManager to make it more convenient to see if a directory exists */

@interface NSFileManager (CVSBundleExtensions)
- (BOOL)directoryExistsAtPath:(NSString *)path;
@end

@implementation NSFileManager (CVSBundleExtensions)
- (BOOL)directoryExistsAtPath:(NSString *)path;
{
    BOOL isDirectory;

    if (![self fileExistsAtPath: path isDirectory: &isDirectory])
        return NO;
    return isDirectory;
}

@end


/*
   On +load we register for notification when a document is saved. When we get that notification, we try to copy the CVS info from the model's backup to the newly saved model.
*/

@interface CVSManager (Private)
+ (NSString *)restoreCVSInfoToDocumentAt:(NSString *)path;
+ (void)didSaveDocument:(NSNotification *)notification;
@end


@implementation CVSManager

static NSString *versionString = @"version 1.3";

+ (NSString *)versionString;	// in case we ever decide to display this somehow
{
    return versionString;
}

+ (void)load;
{
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(didSaveDocument:) name:EOMDocumentDidSaveNotification object:nil];
}

+ (NSString *)adjustEntriesFileInDirectory:(NSString *)basePath forModel:(EOModel *)aModel
//+ (NSString *)adjustEntriesFile:(NSString *)path forModel:(EOModel *)aModel
{
    NSMutableSet *entityNames, *fetchSpecNames, *procedureNames;
    NSMutableString *result;
    NSArray *entriesLines;
    NSEnumerator *enumerator, *remainder = nil;
    NSString *line, *remainderLine = nil;
    NSString *name;
    NSString *cvsBasePath;
    NSString *path;
    NSString *cvsTag;
    NSFileManager *fileManager;

    fileManager = [NSFileManager defaultManager];

    cvsBasePath = [basePath stringByAppendingPathComponent:@"CVS"];

    path = [cvsBasePath stringByAppendingPathComponent:@"Tag"];
    cvsTag = [NSString stringWithContentsOfFile:path];
    if (!cvsTag)
        cvsTag = @"";
    
    else {	// we only want the stuff before the newline
        NSRange range;

        range = [cvsTag rangeOfString:@"\n"];
        if (range.length > 0) {
            range.length = range.location;
            range.location = 0;
            cvsTag = [cvsTag substringWithRange:range];
        }
    }
    
    
    path = [cvsBasePath stringByAppendingPathComponent:@"Entries"];

    result = [NSMutableString string];
    entityNames = [NSMutableSet setWithArray:[aModel entityNames]];
    procedureNames = [NSMutableSet setWithArray:[aModel storedProcedureNames]];
    entriesLines = [[NSString stringWithContentsOfFile:path] componentsSeparatedByString:@"\n"];

    // these aren't really fetch spec names, but names of entities with fetch specs
    fetchSpecNames = [NSMutableSet setWithArray:[aModel entityNames]];
    // remove any fetch spec names for which we don't have files
    enumerator = [entityNames objectEnumerator];
    while ((name = [enumerator nextObject])) {
        NSString *fetchSpecPath;

        fetchSpecPath = [[basePath stringByAppendingPathComponent:name] stringByAppendingPathExtension:@"fspec"];
        if (![fileManager fileExistsAtPath:fetchSpecPath])
            [fetchSpecNames removeObject:name];
    }

    /* Find removed entities */
    enumerator = [entriesLines objectEnumerator];
    while((line = [enumerator nextObject])) {
        NSArray *parts = [line componentsSeparatedByString:@"/"];

        if ([parts count] == 6) {
            name = [parts objectAtIndex:1];

            // entity
            if ([[name pathExtension] isEqualToString:@"plist"]) {
                NSString *version;

                version = [parts objectAtIndex:2];
                name = [name stringByDeletingPathExtension];
                
                if ([entityNames member:name]) {
                    // we have an entity in the Entries file; make sure it's not marked for CVS removal
                    [entityNames removeObject:name];
                    if ([version characterAtIndex:0] == '-') {
                        NSMutableArray *mutable = [parts mutableCopy];

                        [mutable replaceObjectAtIndex:2 withObject:[version substringFromIndex:1]];
                        line = [mutable componentsJoinedByString:@"/"];
                        [mutable release];
                    }
                } else {
                    // we have an entry in the Entries file with no corresponding entity; make sure it's marked for CVS removal
                    if ([version isEqualToString:@"0"])
                        line = nil;
                    else if ([version characterAtIndex:0] != '-') {
                        NSMutableArray *mutable = [parts mutableCopy];

                        [mutable replaceObjectAtIndex:2 withObject:[@"-" stringByAppendingString:version]];
                        line = [mutable componentsJoinedByString:@"/"];
                        [mutable release];
                    }
                }

            // fetch specifications
            } else if ([[name pathExtension] isEqualToString:@"fspec"]) {
                name = [name stringByDeletingPathExtension];
                if ([fetchSpecNames member:name])
                    [fetchSpecNames removeObject:name];
                else {
                    NSString *version = [parts objectAtIndex:2];

                    if ([version isEqualToString:@"0"]) {
//                        NSLog (@"Deleting CVS entry for %@.fspec (was marked for addition but was never committed and has now been removed)", name);
                        line = nil;
                    } else {
                        NSMutableArray *mutable = [parts mutableCopy];

//                        NSLog (@"Marking %@.fspec for removal from CVS", name);
                        [mutable replaceObjectAtIndex:2 withObject:[@"-" stringByAppendingString:version]];
                        line = [mutable componentsJoinedByString:@"/"];
                        [mutable release];
                    }
                }
                
            // stored procedures
            } else if ([[name pathExtension] isEqualToString:@"storedProcedure"]) {
                NSString *version;

                version = [parts objectAtIndex:2];
                name = [name stringByDeletingPathExtension];
                
                if ([procedureNames member:name]) {
                    // we have a stored procedure in the Entries file; make sure it's not marked for CVS removal
                    [procedureNames removeObject:name];
                    if ([version characterAtIndex:0] == '-') {
                        NSMutableArray *mutable = [parts mutableCopy];

                        [mutable replaceObjectAtIndex:2 withObject:[version substringFromIndex:1]];
                        line = [mutable componentsJoinedByString:@"/"];
                        [mutable release];
                    }
                } else {
                    // we have an entry in the Entries file with no corresponding stored procedure; make sure it's marked for CVS removal
                    if ([version isEqualToString:@"0"])
                        line = nil;
                    else {
                        NSMutableArray *mutable = [parts mutableCopy];

                        [mutable replaceObjectAtIndex:2 withObject:[@"-" stringByAppendingString:version]];
                        line = [mutable componentsJoinedByString:@"/"];
                        [mutable release];
                    }
                }
            }
            
            if (line) {
                [result appendString:line];
                [result appendString:@"\n"];
            }
        } else {
            remainder = enumerator;
            remainderLine = line;
        }
    }

    /* Find added entities */
    enumerator = [entityNames objectEnumerator];
    while((name = [enumerator nextObject])) {
        [result appendFormat:@"/%@.plist/0/Initial %@.plist//%@\n", name, name, cvsTag];
    }

    /* Find added fetch specifications */
    enumerator = [fetchSpecNames objectEnumerator];
    while((name = [enumerator nextObject])) {
        [result appendFormat:@"/%@.fspec/0/Initial %@.fspec//%@\n", name, name, cvsTag];
    }

    /* Find added procedures */
    enumerator = [procedureNames objectEnumerator];
    while((name = [enumerator nextObject])) {
        [result appendFormat:@"/%@.storedProcedure/0/Initial %@.storedProcedure//%@\n", name, name, cvsTag];
    }

    /* Append on end */
    if (remainderLine) {
        do {
            [result appendString:remainderLine];
        } while((remainderLine = [remainder nextObject]));
    }
    [result writeToFile:path atomically:YES];
    return nil;
}

+ (NSString *)restoreCVSInfoToDocumentAt:(NSString *)path forModel:(EOModel *)aModel;
{
    NSFileManager *manager;
    NSString *cvsPath;
    NSString *oldPath;
    NSString *basePath;

    basePath = [path copy];

    manager = [NSFileManager defaultManager];

    if (![manager directoryExistsAtPath:path])
        return [NSString stringWithFormat:@"Could not find model at %@", path];

    oldPath = [path stringByAppendingString:@"~"];
    if (![manager directoryExistsAtPath:oldPath])
        return nil;	// backup won't exist if model being saved was brand new

    cvsPath = [oldPath stringByAppendingPathComponent:@"CVS"];
    if (![manager directoryExistsAtPath:cvsPath])
        return nil;	// no CVS info to copy - probably never been checked in

    path = [path stringByAppendingPathComponent:@"CVS"];
    if ([manager directoryExistsAtPath:path])
        return [NSString stringWithFormat:@"CVS directory already exists at %@", path];

    if (![manager movePath:cvsPath toPath:path handler:nil])
        return [NSString stringWithFormat:@"Failed to move '%@' to '%@'", cvsPath, path];

    return [self adjustEntriesFileInDirectory:basePath forModel:aModel];
}

+ (void)didSaveDocument:(NSNotification *)notification;
{
    NSString *errorMsg;
    NSString *path;

    path = [(EOModelerDocument *)[notification object] documentPath];

    errorMsg = [self restoreCVSInfoToDocumentAt:path forModel:[(EOModelerDocument *)[notification object] model]];
    if (errorMsg)
        NSRunAlertPanel (nil, errorMsg, nil, nil, nil);
}

@end
